/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_SUBSTITUTION_HH_
#define HATP_SUBSTITUTION_HH_

#include<iostream>
#include<string>
#include<map>

class Substitution
{
private:
	/**
	* the names of variables associated to a boolean (true if the variable has been instantied)
	* and a string containing its value
	*/
	std::map<std::string, std::pair<bool, std::string> > variables;

public:
	explicit Substitution(std::vector<std::string> const& argVarNames);
	explicit Substitution(Substitution const& sub);
	std::map<std::string, std::pair<bool, std::string> > const& getVariables() const;
	bool isInstancied(std::string const& name) const;
	std::string getValue(std::string const& name) const;
	void setValue(std::string const& name, std::string const& val);
	bool operator==(Substitution const& other) const;
	std::string operator[](const char* name) const;
	~Substitution();
};

/**
 * constructor: explore varNamesAndTypes and complete varNames accordingly
 * for each value in varNamesAndTypes, add a "false" in instantied and an empty string in values
 */
inline Substitution::Substitution(std::vector<std::string> const& argVarNames)
{
	for(unsigned int i=0 ; i<argVarNames.size() ; ++i)
		variables[argVarNames[i]] = std::make_pair(false, "");
}

/**
 * constructor by copying
 */
inline Substitution::Substitution(Substitution const& sub)
{
	variables = sub.getVariables();
}

/**
 * return varNames
 */
inline std::map<std::string, std::pair<bool, std::string> > const& Substitution::getVariables() const
{
	return variables;
}

/**
 * explore varNames to find name and return corresponding value in instantied
 */
inline bool Substitution::isInstancied(std::string const& name) const
{
	std::map<std::string, std::pair<bool, std::string> >::const_iterator it = variables.find(name);

	if(it != variables.end())
		return (it->second).first;

	std::cerr << "ERROR: trying to find a variable called " << name << " in a Substitution, variable unknown (return false)" << std::endl;
	return false;
}

/**
 * explore varNames to find name and return corresponding value in values
 */
inline std::string Substitution::getValue(std::string const& name) const
{
	std::map<std::string, std::pair<bool, std::string> >::const_iterator it = variables.find(name);

	if(it != variables.end())
		return (it->second).second;

	std::cerr << "ERROR: trying to find a variable called " << name << " in a Substitution, variable unknown (return empty string)" << std::endl;
	return std::string("");
}

/**
 * explore varNames to find name and set corresponding value in instantied/values to false/val
 */
inline void Substitution::setValue(std::string const& name, std::string const& val)
{
	std::map<std::string, std::pair<bool, std::string> >::iterator it = variables.find(name);

	if(it != variables.end())
	{
		it->second = std::make_pair(true,val);
		return;
	}

	std::cerr << "ERROR: trying to find a variable called " << name << " in a Substitution, variable unknown (do nothing)" << std::endl;
}

/**
 * return true if other is the same as current Substitution
 */
inline bool Substitution::operator==(Substitution const& other) const
{
	std::map<std::string, std::pair<bool, std::string> > const& otherVar = other.getVariables();
	std::map<std::string, std::pair<bool, std::string> >::const_iterator it, it1;

	if(variables.size() == otherVar.size())
	{
		for(it = otherVar.begin() ; it != otherVar.end() ; ++it)
		{
			it1 = variables.find(it->first);
			if(it1 == variables.end() || it->second != it1->second)
				return false;
		}

		return true;
	}

	return false;
}

/**
 * return a copy of value associated to name
 * if value is not instancied return an empty string
 */
inline std::string Substitution::operator[](const char* name) const
{
	std::string s;

	std::map<std::string, std::pair<bool, std::string> >::const_iterator it = variables.find(name);
	if(it != variables.end() && (it->second).first )
		s = (it->second).second;

	return s;
}

/**
 * destructor: empty
 */
inline Substitution::~Substitution()
{
}

#endif
