/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#if !defined HATP_ATTRIBUTE_SET_HH_
#define HATP_ATTRIBUTE_SET_HH_

#include<string>
#include<list>
#include<set>
#include<vector>
#include<map>
#include<sstream>

#include "HATPexception.hh"

/** Represent all the attributes of the type T.
 * The attributes can be static or dynamic and atom or a set.
 */
template <typename T>
class AttributeSet
{
private:

	/** Default value for atom attributes.
	*/
	T defaultAtomValue;

	/** Default value for set attributes.
	*/
	std::set<T> defaultSetValue;

	/** The static atom attributes.
	*/
	std::map<std::string, T> staticAtomAttrs;

	/** The dynamic atom attributes.
	*/
	std::map<std::string, T> dynamicAtomAttrs;

	/** The static set attributes.
	*/
	std::map<std::string, std::set<T> > staticSetAttrs;

	/** The dynamic set attributes.
	*/
	std::map<std::string, std::set<T> > dynamicSetAttrs;

public:

	AttributeSet(T argDefaultValue);
	explicit AttributeSet(T argDefaultValue, std::vector<std::string> const& argStaticAtomAttrs,
		std::vector<std::string> const& argDynamicAtomAttrs,
		std::vector<std::string> const& argStaticSetAttrs,
		std::vector<std::string> const& argDynamicSetAttrs);
	AttributeSet(AttributeSet<T> const& attrSet);
	T& getDefaultAtomValue();
	T const& getDefaultAtomValue() const;
	std::set<T>& getDefaultSetValue();
	std::set<T> const& getDefaultSetValue() const;
	std::map<std::string, T> const& getStaticAtomAttrs() const;
	std::map<std::string, T>* getPointerOnStaticAtomAttrs() const;
	std::map<std::string, T> const& getDynamicAtomAttrs() const;
	std::map<std::string, std::set<T> > const& getStaticSetAttrs() const;
	std::map<std::string, std::set<T> >* getPointerOnStaticSetAttrs() const;
	std::map<std::string, std::set<T> > const& getDynamicSetAttrs() const;
	T& getAtom(std::string const& attrName);
	const T& getAtom(std::string const& attrName) const;
	std::set<T>& getSet(std::string const& attrName);
	const std::set<T>& getSet(std::string const& attrName) const;
	bool hasAttributeOfName(std::string const& attrName) const;
	bool hasAtomOfName(std::string const& attrName) const;
	bool hasSetOfName(std::string const& attrName) const;
	bool isStatic(std::string const& attrName) const;
	bool isDynamic(std::string const& attrName) const;
	bool isAtom(std::string const& attrName) const;
	bool isSet(std::string const& attrName) const;
	void updateAttrs(std::map<T,T> const& NewValues);
	void getAllAttributeNames(std::list<std::string>& attrList) const;
	void toStringStream(std::ostringstream& st) const;
	void clearDynamicAtomAttrs();
	void clearDynamicSetAttrs();
	bool empty() const;
	~AttributeSet();
};

/** Constructor: empty, create an empty AttributeSet.
 * Initialize defaultAtomValue to argDefaultValue and defaultSetValue to an empty set.
 * @param argDefaultValue The default value for the atom attributes of type T
 */
template<typename T> inline AttributeSet<T>::AttributeSet(T argDefaultValue) : defaultAtomValue(argDefaultValue) {
	defaultSetValue = std::set<T>();
}

/** Constructor: store data inside arguments depending on their type static/dynamic and atom/set.
 * The lists given as argument contain the name of the attributes, all will be initialised to the default value.
 * @param argDefaultValue The defalut value for the atom attributes of type T
 * @param argStaticAtomAttrs Initialise the static-atom-attribute list
 * @param argDynamicAtomAttrs Initialise the dynamic-atom-attribute list
 * @param argStaticSetAttrs Initialise the static-set-attribute list
 * @param argDynamicSetAttrs Initialise the dynamic-set-attribute list
 */
template<typename T> inline AttributeSet<T>::AttributeSet(T argDefaultValue, std::vector<std::string> const& argStaticAtomAttrs,
	std::vector<std::string> const& argDynamicAtomAttrs,
	std::vector<std::string> const& argStaticSetAttrs,
	std::vector<std::string> const& argDynamicSetAttrs)
: defaultAtomValue(argDefaultValue) {
	defaultSetValue = std::set<T>();
	size_t i;

	for(i=0 ; i<argStaticAtomAttrs.size() ; ++i)
		staticAtomAttrs[argStaticAtomAttrs[i]] = defaultAtomValue;

	for(i=0 ; i<argDynamicAtomAttrs.size() ; ++i)
		dynamicAtomAttrs[argDynamicAtomAttrs[i]] = defaultAtomValue;

	for(i=0 ; i<argStaticSetAttrs.size() ; ++i)
		staticSetAttrs[argStaticSetAttrs[i]] = defaultSetValue;

	for(i=0 ; i<argDynamicSetAttrs.size() ; ++i)
		dynamicSetAttrs[argDynamicSetAttrs[i]] = defaultSetValue;
}

/** Copy constructor.
 * @param attrSet AttributeSet to copy
 */
template<typename T> inline AttributeSet<T>::AttributeSet(AttributeSet<T> const& attrSet) {
	defaultAtomValue = attrSet.getDefaultAtomValue();
	defaultSetValue = attrSet.getDefaultSetValue();
	staticAtomAttrs = attrSet.getStaticAtomAttrs();
	dynamicAtomAttrs = attrSet.getDynamicAtomAttrs();
	staticSetAttrs = attrSet.getStaticSetAttrs();
	dynamicSetAttrs = attrSet.getDynamicSetAttrs();
}

/** Return the default value for atom attributes.
 * @return Default atom value
 */
template<typename T> inline T& AttributeSet<T>::getDefaultAtomValue() {
	return defaultAtomValue;
}

/** Return the default value for atom attributes (const version).
 * @return Default atom value
 */
template<typename T> inline T const& AttributeSet<T>::getDefaultAtomValue() const {
	return defaultAtomValue;
}

/** Return the default set for an attribute.
 * @return Default set value
 */
template<typename T> inline std::set<T>& AttributeSet<T>::getDefaultSetValue() {
	return defaultSetValue;
}

/** Return the default set for an attribute (const version).
 * @return DefaultSetValue
 */
template<typename T> inline std::set<T> const& AttributeSet<T>::getDefaultSetValue() const {
	return defaultSetValue;
}

/** Return the set of static atom attributes and their value.
 * @return Static atom attributes and their value
 */
template<typename T> inline std::map<std::string, T> const& AttributeSet<T>::getStaticAtomAttrs() const {
	return staticAtomAttrs;
}

/** Return a pointer to the set of static atom attributes and their value.
 * @return Pointer to the static atom attributes and their value
 */
template<typename T> inline std::map<std::string, T>* AttributeSet<T>::getPointerOnStaticAtomAttrs() const {
	return const_cast< std::map<std::string, T>* >(&staticAtomAttrs);
}

/** Return the set of dynamic atom attributes and their value.
 * @return Dynamic atom attributes and their value
 */
template<typename T> inline std::map<std::string, T> const& AttributeSet<T>::getDynamicAtomAttrs() const {
	return dynamicAtomAttrs;
}

/** Return the set of static set attributes and their value.
 * @return Static set attributes and their value
 */
template<typename T> inline std::map<std::string, std::set<T> > const& AttributeSet<T>::getStaticSetAttrs() const {
	return staticSetAttrs;
}

/** Return a pointer to the set of static set attributes and their value.
 * @return Pointer to the static set attributes and their value
 */
template<typename T> inline std::map<std::string, std::set<T> >* AttributeSet<T>::getPointerOnStaticSetAttrs() const {
	return const_cast < std::map<std::string, std::set<T> >* > (&staticSetAttrs);
}

/** Return the set of dynamic set attributes and their value.
 * @return Dynamic set attributes and their value
 */
template<typename T> inline std::map<std::string, std::set<T> > const& AttributeSet<T>::getDynamicSetAttrs() const {
	return dynamicSetAttrs;
}

/** Return a reference to the value of the atom attribute attrName.
 * @param attrName Name of the atom attribute
 * @return Reference on the value of attrName
 * @throw Throw an HATPexception if attrName is not the name of an existing attribute
 */
template<typename T> inline T& AttributeSet<T>::getAtom(std::string const& attrName) {
	typename std::map<std::string, T>::iterator it;

	it = staticAtomAttrs.find(attrName);
	if(it != staticAtomAttrs.end())
		return it->second;

	it = dynamicAtomAttrs.find(attrName);
	if(it != dynamicAtomAttrs.end())
		return it->second;

	throw HATPexception(ERROR,"trying to access to a non existing attribute "+attrName);
	return defaultAtomValue;
}

/** Return a const reference to the value of the atom attribute attrName.
 * @param attrName Name of the atom attribute
 * @return Const reference on the value of attrName
 * @throw Throw an HATPexception if attrName is not the name of an existing attribute
 */
template<typename T> inline T const& AttributeSet<T>::getAtom(std::string const& attrName) const {
	typename std::map<std::string, T>::const_iterator it;

	it = staticAtomAttrs.find(attrName);
	if(it != staticAtomAttrs.end())
		return it->second;

	it = dynamicAtomAttrs.find(attrName);
	if(it != dynamicAtomAttrs.end())
		return it->second;

	throw HATPexception(ERROR,"trying to get the value of a non existing attribute "+attrName);
	return defaultAtomValue;
}

/** Return a reference to the value of the set attribute attrName.
 * @param attrName Name of the set attribute
 * @return Reference on the value of attrName
 * @throw Throw an HATPexception if attrName is not the name of an existing attribute
 */
template<typename T> inline std::set<T>& AttributeSet<T>::getSet(std::string const& attrName) {
	typename std::map<std::string, std::set<T> >::iterator it;

	it = staticSetAttrs.find(attrName);
	if(it != staticSetAttrs.end())
		return it->second;

	it = dynamicSetAttrs.find(attrName);
	if(dynamicSetAttrs.find(attrName) != dynamicSetAttrs.end())
		return it->second;

	throw HATPexception(ERROR,"trying to access to a non existing attribute "+attrName);
	return defaultSetValue;
}

/** Return a const reference to the value of the set attribute attrName.
 * @param attrName Name of the set attribute
 * @return Const reference on the value of attrName
 * @throw Throw an HATPexception if attrName is not the name of an existing attribute
 */
template<typename T> inline std::set<T> const& AttributeSet<T>::getSet(std::string const& attrName) const {
	typename std::map<std::string, std::set<T> >::const_iterator it;

	it = staticSetAttrs.find(attrName);
	if(it != staticSetAttrs.end())
		return it->second;

	it = dynamicSetAttrs.find(attrName);
	if(dynamicSetAttrs.find(attrName) != dynamicSetAttrs.end())
		return it->second;

	throw HATPexception(ERROR,"trying to access to a non existing attribute "+attrName);
	return defaultSetValue;
}

/** Check if an attribute with the name attrName exists (either as a static/dynamic and atom/set).
 * @param attrName The attribute name to check
 * @return True if there is an attribute named attrName
 */
template<typename T> inline bool AttributeSet<T>::hasAttributeOfName(std::string const& attrName) const {
	return staticAtomAttrs.find(attrName) != staticAtomAttrs.end() ||
	dynamicAtomAttrs.find(attrName) != dynamicAtomAttrs.end() ||
	staticSetAttrs.find(attrName) != staticSetAttrs.end() ||
	dynamicSetAttrs.find(attrName) != dynamicSetAttrs.end();
}

/** Check if an atom attribute exists with the name attrName (either static or dynamic).
 * @param attrName The attribute name to check
 * @return True if an attribute exits
 */
template<typename T> inline bool AttributeSet<T>::hasAtomOfName(std::string const& attrName) const {
	return staticAtomAttrs.find(attrName) != staticAtomAttrs.end() ||
	dynamicAtomAttrs.find(attrName) != dynamicAtomAttrs.end();
}

/** Check if a set attribute exists with the name attrName (either static or dynamic).
 * @param attrName The attribute name to check
 * @return True if an attribute exits
 */
template<typename T> inline bool AttributeSet<T>::hasSetOfName(std::string const& attrName) const {
	return staticSetAttrs.find(attrName) != staticSetAttrs.end() ||
	dynamicSetAttrs.find(attrName) != dynamicSetAttrs.end();
}

/** Check if the attribute attrName is either static or not (the attribute can be an atom or set attribute).
 * @param attrName The attribute the check
 * @return True if the attribute is static
 */
template<typename T> inline bool AttributeSet<T>::isStatic(std::string const& attrName) const {
	return staticAtomAttrs.find(attrName) != staticAtomAttrs.end() ||
	staticSetAttrs.find(attrName) != staticSetAttrs.end();
}

/** Check if the attribute attrName is either dynamic or not (the attribute can be an atom or set attribute).
 * @param attrName The attribute the check
 * @return True if the attribute is dynamic
 */
template<typename T> inline bool AttributeSet<T>::isDynamic(std::string const& attrName) const {
	return dynamicAtomAttrs.find(attrName) != dynamicAtomAttrs.end() ||
	dynamicSetAttrs.find(attrName) != dynamicSetAttrs.end();
}

/** Check if the attribute attrName is either an atom attribute or not (the attribute can be static or dynamic).
 * @param attrName The attribute the check
 * @return True if the attribute is an atom attribute
 */
template<typename T> inline bool AttributeSet<T>::isAtom(std::string const& attrName) const {
	return staticAtomAttrs.find(attrName) != staticAtomAttrs.end() ||
	dynamicAtomAttrs.find(attrName) != dynamicAtomAttrs.end();
}

/** Check if the attribute attrName is either a set attribute or not (the attribute can be static or dynamic).
 * @param attrName The attribute the check
 * @return True if the attribute is a set attribute
 */
template<typename T> inline bool AttributeSet<T>::isSet(std::string const& attrName) const {
	return staticSetAttrs.find(attrName) != staticSetAttrs.end() ||
	dynamicSetAttrs.find(attrName) != dynamicSetAttrs.end();
}

/** Update the values of attributes.
 * Explores all attributes and when one has a value NewValues->first, replace this value by NewValues->second.
 * @param NewValues A map where the keys are the old values and the content for a key is the new value
 */
template<typename T> inline void AttributeSet<T>::updateAttrs(std::map<T,T> const& NewValues) {
	typename std::map<std::string, T>::iterator it1;
	typename std::map<std::string, std::set<T> >::iterator it2;
	typename std::map<T, T>::const_iterator itVal;
	typename std::set<T>::iterator itSet;
	for(itVal = NewValues.begin() ; itVal != NewValues.end() ; ++itVal) {
		for(it1=staticAtomAttrs.begin() ; it1 != staticAtomAttrs.end() ; ++it1) 
			if(it1->second == itVal->first) it1->second = itVal->second;
		for(it1=dynamicAtomAttrs.begin() ; it1 != dynamicAtomAttrs.end() ; ++it1) 
			if(it1->second == itVal->first) it1->second = itVal->second;
		for(it2=staticSetAttrs.begin() ; it2 != staticSetAttrs.end() ; ++it2) {
			for(itSet = (it2->second).begin() ; itSet != (it2->second).end() ; ++itSet)
				if(*itSet == itVal->first) {
					(it2->second).erase(itSet);
					(it2->second).insert(itVal->second);
					break;
				}
		}
		for(it2=dynamicSetAttrs.begin() ; it2 != dynamicSetAttrs.end() ; ++it2) {
			for(itSet = (it2->second).begin() ; itSet != (it2->second).end() ; ++itSet)
				if(*itSet == itVal->first) {
					(it2->second).erase(itSet);
					(it2->second).insert(itVal->second);
					break;
				}
		}
	}
}

/** Return the list of all attributes (static and dynamic, atom and set).
 * Complete attrList with attribute names of the AttributeSet.
 * @param attrList The list that is filled by the all the attributes' name
 */
template<typename T> inline void AttributeSet<T>::getAllAttributeNames(std::list<std::string>& attrList) const {
	typename std::map<std::string, T>::const_iterator it1;
	typename std::map<std::string, std::set<T> >::const_iterator it2;

	for(it1=staticAtomAttrs.begin() ; it1 != staticAtomAttrs.end() ; ++it1)
		attrList.push_back(it1->first);

	for(it1=dynamicAtomAttrs.begin() ; it1 != dynamicAtomAttrs.end() ; ++it1)
		attrList.push_back(it1->first);

	for(it2=staticSetAttrs.begin() ; it2 != staticSetAttrs.end() ; ++it2)
		attrList.push_back(it2->first);

	for(it2=dynamicSetAttrs.begin() ; it2 != dynamicSetAttrs.end() ; ++it2)
		attrList.push_back(it2->first);
}

/** Empty the dynamic atom attribute list.
 */
template<typename T> inline void AttributeSet<T>::clearDynamicAtomAttrs() {
	using namespace std;
	std::string tmp;
	typename std::map<std::string, T>::iterator it1;
	for(it1=dynamicAtomAttrs.begin() ; it1 != dynamicAtomAttrs.end() ; ++it1)
		dynamicAtomAttrs[it1->first] = T();
}

/** Empty the dynamic set attribute list.
 */
template<typename T> inline void AttributeSet<T>::clearDynamicSetAttrs() {
	typename std::set<T>::iterator itSet;
	typename std::map<std::string, std::set<T> >::iterator it2;
	for(it2=dynamicSetAttrs.begin() ; it2 != dynamicSetAttrs.end() ; ++it2)
		for(itSet = (it2->second).begin() ; itSet != (it2->second).end() ; ++itSet) {
		(it2->second).erase(itSet);
	}
}

/** Print all the attributes to the stream givean as argument.
 * Print the attributes, their type and their value. Like :
 * - name : value (static)
 * - name : value (dynamic)
 * - name : {value, value, ... } (static)
 * - name : {value, value, ... } (dynamic)
 *
 * @param st The stream the attributes will be printed in
 */
template<typename T> inline void AttributeSet<T>::toStringStream(std::ostringstream& st) const {
	typename std::map<std::string, T>::const_iterator it1;
	typename std::map<std::string, std::set<T> >::const_iterator it2;
	typename std::set<T>::const_iterator setIT;
	std::set<T> setData;

	for(it1=staticAtomAttrs.begin() ; it1 != staticAtomAttrs.end() ; ++it1)
		st << "\t\t" << it1->first + ": " << it1->second << " (static)" << std::endl;

	for(it1=dynamicAtomAttrs.begin() ; it1 != dynamicAtomAttrs.end() ; ++it1)
		st << "\t\t" << it1->first + ": " << it1->second << " (dynamic)" << std::endl;

	for(it2=staticSetAttrs.begin() ; it2 != staticSetAttrs.end() ; ++it2)
	{
		setData = it2->second;
		st << "\t\t" << it2->first << ": {";
		for(setIT=setData.begin() ; setIT != setData.end() ;)
		{
			st << *setIT;
			if(++setIT != setData.end())
				st << ", ";
		}
		st << "} (static)" << std::endl;
	}

	for(it2=dynamicSetAttrs.begin() ; it2 != dynamicSetAttrs.end() ; ++it2)
	{
		setData = it2->second;
		st << "\t\t" << it2->first << ": {";
		for(setIT=setData.begin() ; setIT != setData.end() ;)
		{
			st << *setIT;
			if(++setIT != setData.end())
				st << ", ";
		}
		st << "} (dynamic)" << std::endl;
	}
}

/** Specialised version of toStringStream for bool attributes.
 * See AttributeSet<T>::toStringStream(std::ostringstream& st)
 * @param st The stream the attributes will be printed in
 */
template<> inline void AttributeSet<bool>::toStringStream(std::ostringstream& st) const {
	std::map<std::string, bool>::const_iterator it1;
	std::map<std::string, std::set<bool> >::const_iterator it2;
	std::set<bool>::const_iterator setIT;
	std::set<bool> setData;
	std::string sTrue("true"), sFalse("false");
	for(it1=staticAtomAttrs.begin() ; it1 != staticAtomAttrs.end() ; ++it1) {
		st << "\t\t" << it1->first + ": ";
		if(it1->second) st << sTrue;
		else st << sFalse;
		st << " (static)" << std::endl;
	}
	for(it1=dynamicAtomAttrs.begin() ; it1 != dynamicAtomAttrs.end() ; ++it1) {
		st << "\t\t" << it1->first + ": ";
		if(it1->second) st << sTrue;
		else st << sFalse;
		st << " (dynamic)" << std::endl;
	}
	for(it2=staticSetAttrs.begin() ; it2 != staticSetAttrs.end() ; ++it2) {
		setData = it2->second;
		st << "\t\t" << it2->first << ": {";
		for(setIT=setData.begin() ; setIT != setData.end() ;) {
			if(*setIT) st << sTrue;
			else st << sFalse;
			if(++setIT != setData.end()) st << ", ";
		}
		st << "} (static)" << std::endl;
	}
	for(it2=dynamicSetAttrs.begin() ; it2 != dynamicSetAttrs.end() ; ++it2) {
		setData = it2->second;
		st << "\t\t" << it2->first << ": {";
		for(setIT=setData.begin() ; setIT != setData.end() ;) {
			if(*setIT) st << sTrue;
			else st << sFalse;
			if(++setIT != setData.end()) st << ", ";
		}
		st << "} (dynamic)" << std::endl;
	}
}

/** Check if all attribute list are empty.
 * @return True if the AttributeSet is empty
 */
template<typename T> inline bool AttributeSet<T>::empty() const {
	return staticAtomAttrs.empty() && dynamicAtomAttrs.empty() &&
	staticSetAttrs.empty() && dynamicSetAttrs.empty();
}


/** Destructor: empty
 */
template<typename T> inline AttributeSet<T>::~AttributeSet() {}

#endif
